// $Id: kagura4.c 43 2009-07-07 12:34:29Z takanori $

/* Declarations for kagura4.c.
   Copyright (C) 2009 Yamamoto Takanori / Tatyana.

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include <sys/param.h>     // MAXPATHLEN
#include <dirent.h>        // MAXNAMLEN
#include <sys/stat.h>      // mkdir()

struct header_pak {
  u_int8_t brand[4]; u_int16_t v0, v1; u_int8_t un0[8];
  u_int32_t v2,v3,v4,extracted_data_size;
  u_int32_t compressed_data_size, v7;
};

struct header_item {
  u_int32_t pos, v0, extracted_data_size, v1,
    compressed_data_size, v2, type, v3, v4,
    v5, v6, v7, v8;
};

u_int32_t l2b_32 (u_int32_t n) {
  return
    ((n>>24) & 0x000000ff) |
    ((n>> 8) & 0x0000ff00) |
    ((n<< 8) & 0x00ff0000) |
    ((n<<24) & 0xff000000);
}

int readHeader_pak (FILE **fp, struct header_pak *p){
  if (fread(p, sizeof(struct header_pak), 1, *fp) != 1) {
    return -1;
  }
  if (strncmp((const char*)p->brand, "PAK", 4)) {
    return -1;
  }
#if defined (__ppc__) || defined (__ppc64__)
  p->extracted_data_size = l2b_32(p->extracted_data_size);
  p->compressed_data_size = l2b_32(p->compressed_data_size);
#endif
  return 0;
}

int getHeader_item (unsigned char **src, size_t *src_len, struct header_item *p, char fn[]) {
  if (*src_len <= sizeof(struct header_item)) {
    return -1;
  }
  memcpy(p, *src, sizeof(struct header_item));
#if defined (__ppc__) || defined (__ppc64__)
  p->pos = l2b_32(p->pos);
  p->extracted_data_size = l2b_32(p->extracted_data_size);
  p->compressed_data_size = l2b_32(p->compressed_data_size);
  p->type = l2b_32(p->type);
#endif
  unsigned char *q = (*src) + sizeof(struct header_item);
  while (*q) {
    ++q;
    if (q>=(*src)+*src_len) {
      return -1;
    }
  }
  strlcpy(fn, (const char *)(*src) + sizeof(struct header_item), MAXNAMLEN);
  *src = q+1;
  (*src_len) -= (sizeof(struct header_item) + strlen(fn) + 1);
  return 0;
}

int inflateIt(unsigned char *compressed_data, size_t compressed_data_size, unsigned char *extracted_data, size_t extracted_data_size) {
  if (compressed_data_size == 0) {return 0;} // workaround for ft_05h.txt in mesh.pak

  z_stream z;
  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  z.next_in = Z_NULL;
  z.avail_in = 0;

  if (inflateInit2(&z, -MAX_WBITS) != Z_OK) { // raw inflate mode
    return -1;
  }

  z.next_in = compressed_data;
  z.avail_in = compressed_data_size;
  z.next_out = extracted_data;
  z.avail_out = extracted_data_size;
  int ret = inflate(&z, Z_NO_FLUSH);
  inflateEnd(&z);

  if (ret != Z_STREAM_END) {
    return -1;
  }
  return 0;
}

int main(int argc, char *argv[]) {
  if (argc==1) {
    fprintf(stderr, "usage : %s [PAK file]\n", argv[0]);
    return 0;
  }

  FILE *fp_src;
  if ((fp_src = fopen(argv[1], "rb")) == NULL) {
    fprintf(stderr, "Failed to open PAK file.\n");
    return -1;
  }

  struct header_pak h0;
  if (readHeader_pak(&fp_src, &h0)) {
    fprintf(stderr, "Failed to readHeader_pak().\n");
    return -1;
  }
  
  fprintf(stderr, "EXTRACTED SIZE  : 0x%08x\n", h0.extracted_data_size);
  fprintf(stderr, "COMPRESSED SIZE : 0x%08x\n", h0.compressed_data_size);

  // 

  unsigned char *extracted_data, *compressed_data;
  if ((extracted_data = (unsigned char *)malloc(h0.extracted_data_size)) == NULL) {
    fprintf(stderr, "Failed to allocate extracted_data.\n");
    return -1;
  }
  if ((compressed_data = (unsigned char *)malloc(h0.compressed_data_size)) == NULL) {
    fprintf(stderr, "Failed to allocate compressed_data.\n");
    return -1;
  }

  fseek(fp_src, sizeof(struct header_pak), SEEK_SET);
  fread(compressed_data, 1, h0.compressed_data_size, fp_src);

  if (inflateIt(compressed_data, h0.compressed_data_size, extracted_data, h0.extracted_data_size)) {
    fprintf(stderr, "Failed to inflate.\n");
    return -1;
  }

  struct header_item h1;
  char dir1[MAXPATHLEN] = "";
  char fn1[MAXNAMLEN];
  unsigned char *p = extracted_data;
  size_t plen = h0.extracted_data_size;

  int mode=0, depth=0;
  while (plen) {
    if (getHeader_item(&p, &plen, &h1, fn1)) {break;}
    if (h1.type == 0x00000010) { //DIR
      if (mode==2) {
        while (depth>1) {
          strlcat(dir1, "../", MAXPATHLEN);
          --depth;
        }
      }
      strlcat(dir1, fn1, MAXPATHLEN);
      strlcat(dir1, "/", MAXPATHLEN);
      depth++;
      mode=1;
      mkdir(dir1, 0755);
    } else { //FILE
      mode=2;
      unsigned char *extracted_item, *compressed_item;
      if ((extracted_item = (unsigned char *)malloc(h1.extracted_data_size)) == NULL) {
        fprintf(stderr, "Failed to allocate extracted_item.\n");
        return -1;
      }
      if ((compressed_item = (unsigned char *)malloc(h1.compressed_data_size)) == NULL) {
        fprintf(stderr, "Failed to allocate compressed_item.\n");
        return -1;
      }
      fseek(fp_src, sizeof(struct header_pak)+h0.compressed_data_size+h1.pos, SEEK_SET);
      fread(compressed_item, 1, h1.compressed_data_size, fp_src);

#if 1
      fprintf(stderr, "\n");
      fprintf(stderr, "ITEM POSITION        : 0x%08x\n", h1.pos);
      fprintf(stderr, "ITEM EXTRACTED SIZE  : 0x%08x\n", h1.extracted_data_size);
      fprintf(stderr, "ITEM COMPRESSED SIZE : 0x%08x\n", h1.compressed_data_size);
      fprintf(stderr, "ITEM TYPE            : 0x%08x\n", h1.type);
      fprintf(stderr, "ITEM V0              : 0x%08x\n", h1.v0);
      fprintf(stderr, "ITEM V1              : 0x%08x\n", h1.v1);
      fprintf(stderr, "ITEM V2              : 0x%08x\n", h1.v2);
      fprintf(stderr, "ITEM V3              : 0x%08x\n", h1.v3);
      fprintf(stderr, "ITEM V4              : 0x%08x\n", h1.v4);
      fprintf(stderr, "ITEM V5              : 0x%08x\n", h1.v5);
      fprintf(stderr, "ITEM V6              : 0x%08x\n", h1.v6);
      fprintf(stderr, "ITEM V7              : 0x%08x\n", h1.v7);
      fprintf(stderr, "ITEM V8              : 0x%08x\n", h1.v8);
      fprintf(stderr, "ITEM FILENAME        : %s%s\n",   dir1, fn1);
      fprintf(stderr, "PLEN                 : 0x%08x\n", plen);
#endif

      if (inflateIt(compressed_item, h1.compressed_data_size, extracted_item, h1.extracted_data_size)) {
        fprintf(stderr, "Failed to inflate.\n");
        return -1;
      }

      char path1[MAXPATHLEN];
      strlcpy(path1, dir1, MAXPATHLEN);
      strlcat(path1, fn1, MAXPATHLEN);
      FILE *fp_dst;
      if ((fp_dst = fopen(path1, "wb")) == NULL) {
        fprintf(stderr, "Failed to open file.\n");
        return -1;
      }
      fwrite(extracted_item, 1, h1.extracted_data_size, fp_dst);
      fclose(fp_dst);
      free(extracted_item);
      free(compressed_item);
    }
  }
  
  free(extracted_data);
  free(compressed_data);

  fclose(fp_src);

  return 0;
}
